@@ -41,30 +41,45 @@ typedef enum
 	NOTIF_NICK
 } NotifStatus;
 
+/* Represents a channel entry in the pop-up menu. */
 struct MenuChannel
 {
 	NotifStatus	status;
 	GtkWidget*	menu_item;
 };
 
-static xchat_plugin*			ph;							/* Plugin handle. */
-static xchat_gnome_plugin*	xgph;							/* xchat gnome plugin handle. */
-static NotifStatus			status = NOTIF_NONE;		/* Current status level. */
-static gboolean				window_visible = TRUE;	/* Keep track of whether the window is visible. */
-static GtkWidget*				main_window;				/* xchat-gnome's main window. */
-static GHashTable*			channels;					/* A reference to the navigation tree. */
-static EggTrayIcon*			notification;				/* Notification area icon. */
-static GtkMenu*				menu;							/* The menu that pops up. */
-static GtkWidget*				image;						/* The image displayed by the icon. */
-static GdkPixbuf*				pixbufs[4];					/* Pixbufs */
+static xchat_plugin*		ph;			/* Plugin handle. */
+static xchat_gnome_plugin*	xgph;			/* xchat gnome plugin handle. */
+static NotifStatus		status = NOTIF_NONE;	/* Current status level. */
+static gboolean			window_visible = TRUE;	/* Keep track of whether the window is visible. */
+static gboolean			focused = TRUE;		/* GTK_WIDGET_HAS_FOCUS doesn't seem to be working... */
+static gboolean			persistant;		/* Keep the icon in the tray at all times? */
+static GtkWidget*		main_window;		/* xchat-gnome's main window. */
+static GHashTable*		channels;		/* A reference to the navigation tree. */
+static EggTrayIcon*		notification;		/* Notification area icon. */
+static GtkMenu*			menu;			/* The menu that pops up. */
+static GtkWidget*		image;			/* The image displayed by the icon. */
+static GdkPixbuf*		pixbufs[4];		/* Pixbufs */
 
 
 /*** Callbacks ***/
 static gboolean
 got_focus_cb (GtkWidget * widget, GdkEventFocus * event, gpointer data)
 {
-	/* Hide the notification icon. */
-	gtk_widget_hide_all (GTK_WIDGET (notification));
+	focused = TRUE;
+
+	/* Reset the status. */
+	status = NOTIF_NONE;
+
+	if (!persistant) {
+		/* Hide the notification icon. */
+		gtk_widget_hide_all (GTK_WIDGET (notification));
+	} else {
+		gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbufs[0]);
+
+		/* Show the notification icon. */
+		gtk_widget_show_all (GTK_WIDGET (notification));
+	}
 
 	return FALSE;
 }
@@ -72,44 +87,19 @@ got_focus_cb (GtkWidget * widget, GdkEventFocus * event, gpointer data)
 static gboolean
 lost_focus_cb (GtkWidget * widget, GdkEventFocus * event, gpointer data)
 {
-	/* Reset the status. */
-	status = NOTIF_NONE;
-	gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbufs[0]);
-
-	/* Show the notification icon. */
-	gtk_widget_show_all (GTK_WIDGET (notification));
-
+	focused = FALSE;
 	return FALSE;
 }
 
 static int
 new_msg_cb (char **word, void *msg_lvl)
 {
-	gchar*					chan_name = (gchar*) xchat_get_info (ph, "channel");
-	struct MenuChannel*	chan = (struct MenuChannel*) g_hash_table_lookup (channels, (gconstpointer) chan_name);
+	gchar* chan_name = (gchar*) xchat_get_info (ph, "channel");
 
-	if (chan == NULL) {
-		/* If we get a message for a channel we're not already in, add it.
-		 * This takes care of the problems David and I were seeing with our proxies
-		 * and the hash table not being populated if the plugin is loaded when x-g
-		 * starts.
-		 */
-		chan = (struct MenuChannel*) malloc (sizeof (struct MenuChannel));
-		chan->status = (NotifStatus) msg_lvl;
-		chan->menu_item = gtk_menu_item_new_with_label (chan_name);
-		g_hash_table_insert (channels, (gpointer) chan_name, (gpointer) chan);
-	}
-
-	if (chan->status < (NotifStatus) msg_lvl) {
-		chan->status = (NotifStatus) msg_lvl;
-		/* FIXME memory leak? */
-		//gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (chan->menu_item),
-				//gtk_image_new_from_pixbuf (pixbufs[(int) chan->status]));
-	}
-
-	if (status < (NotifStatus) msg_lvl && !GTK_WIDGET_HAS_FOCUS (main_window)) {
+	if (status < (NotifStatus) msg_lvl && !focused) {
 		status = (NotifStatus) msg_lvl;
 		gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbufs[status]);
+		gtk_widget_show_all (GTK_WIDGET (notification));
 	}
 
 	return 0;
@@ -118,24 +108,19 @@ new_msg_cb (char **word, void *msg_lvl)
 static int
 join_chan_cb (char **word, void *data)
 {
-	struct MenuChannel* item = (struct MenuChannel*) malloc (sizeof (struct MenuChannel));
-
-	item->status = NOTIF_NONE;
-	item->menu_item = gtk_menu_item_new_with_label ((gchar*) word[1]);
-	g_hash_table_insert (channels, (gpointer) word[1], (gpointer) item);
-
 	return 0;
 }
 
 static int
 part_chan_cb (char **word, void *data)
 {
-	struct MenuChannel*	chan = (struct MenuChannel*) g_hash_table_lookup (channels, word[2]);
+	return 0;
+}
 
-	//gtk_container_remove (GTK_CONTAINER (menu), GTK_WIDGET (chan->menu_item));
-	g_hash_table_remove (channels, (gconstpointer) word[2]);
-
-	free (chan);
+static int
+chan_changed_cb (char **word, void *data)
+{
+	gchar* chan_name = (gchar*) xchat_get_info (ph, "channel");
 
 	return 0;
 }
@@ -143,7 +128,6 @@ part_chan_cb (char **word, void *data)
 static void
 notification_menu_show (GdkEventButton *event)
 {
-	gtk_menu_popup (menu, NULL, NULL, NULL, NULL, event->button, event->time);
 }
 
 static gboolean
@@ -163,7 +147,7 @@ notification_clicked_cb (GtkWidget * widget, GdkEventButton * event, gpointer da
 
 		/* Right click. */
 	case 3:
-		//notification_menu_show (event);
+		notification_menu_show (event);
 		break;
 
 	default:
@@ -178,15 +162,9 @@ notification_clicked_cb (GtkWidget * widget, GdkEventButton * event, gpointer da
 gboolean
 add_channels_foreach_cb (GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter, gpointer data)
 {
-	struct MenuChannel*	item = (struct MenuChannel*) malloc (sizeof (struct MenuChannel));
-	gchar*					channel;
+	gchar* channel;
 
 	gtk_tree_model_get (model, iter, 1, &channel, -1);
-	item->status = NOTIF_NONE;
-	item->menu_item = gtk_menu_item_new_with_label (channel);
-	g_hash_table_insert (channels, (gpointer) channel, (gpointer) item);
-
-	//gtk_menu_append (menu, item->menu_item);
 
 	return FALSE;
 }
@@ -217,12 +195,6 @@ xchat_gnome_plugin_init (xchat_gnome_plugin * xg_plugin)
 	g_signal_connect (main_window, "focus-out-event", G_CALLBACK (lost_focus_cb), NULL);
 
 	/* Create the menu. */
-	//menu = GTK_MENU (gtk_menu_new ());
-
-	channels = g_hash_table_new (g_str_hash , g_str_equal);
-	gtk_tree_model_foreach (chan_model, add_channels_foreach_cb, NULL);
-
-	//gtk_widget_show (GTK_WIDGET (menu));
 
 	return 1;
 }
@@ -232,14 +204,18 @@ xchat_gnome_plugin_init (xchat_gnome_plugin * xg_plugin)
 int
 xchat_plugin_init (xchat_plugin * plugin_handle, char **plugin_name, char **plugin_desc, char **plugin_version, char *arg)
 {
-	GtkWidget *box;
-	GdkPixbuf *p;
+	GtkWidget*		box;
+	GdkPixbuf*		p;
+	GConfClient*	client = gconf_client_get_default ();
 
 	ph = plugin_handle;
 
 	/* Set the plugin info. */
 	xchat_plugin_get_info (plugin_name, plugin_desc, plugin_version, NULL);
 
+	/* Get our preferences from gconf. */
+	persistant = gconf_client_get_bool (client, "/apps/xchat/plugins/notification/persistant", NULL);
+
 	/* FIXME It would be nice to determine the size of the panel and load these
 	 *       images at that size.
 	 */
@@ -280,14 +256,15 @@ xchat_plugin_init (xchat_plugin * plugin_handle, char **plugin_name, char **plug
 
 
 	/* Hook up our callbacks. */
-	xchat_hook_print (ph, "Channel Notice",            XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_DATA);
-	xchat_hook_print (ph, "Channel Message",           XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
-	xchat_hook_print (ph, "Channel Action",            XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
-	xchat_hook_print (ph, "Channel Msg Hilight",       XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_NICK);
-	xchat_hook_print (ph, "Channel Action Hilight",    XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_NICK);
-	xchat_hook_print (ph, "Private Message to Dialog", XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
-	xchat_hook_print (ph, "You Join",                  XCHAT_PRI_NORM, join_chan_cb, NULL);
-	xchat_hook_print (ph, "You Part",                  XCHAT_PRI_NORM, part_chan_cb, NULL);
+	xchat_hook_print (ph, "Channel Notice",			XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_DATA);
+	xchat_hook_print (ph, "Channel Message",		XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
+	xchat_hook_print (ph, "Channel Action",			XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
+	xchat_hook_print (ph, "Channel Msg Hilight",		XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_NICK);
+	xchat_hook_print (ph, "Channel Action Hilight",		XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_NICK);
+	xchat_hook_print (ph, "Private Message to Dialog",	XCHAT_PRI_NORM, new_msg_cb, (gpointer) NOTIF_MSG);
+	xchat_hook_print (ph, "You Join",			XCHAT_PRI_NORM, join_chan_cb, NULL);
+	xchat_hook_print (ph, "You Part",			XCHAT_PRI_NORM, part_chan_cb, NULL);
+	xchat_hook_print (ph, "Focus Tab",			XCHAT_PRI_NORM, chan_changed_cb, NULL);
 
 	xchat_print (ph, "Notification plugin loaded.\n");
 
@@ -308,8 +285,3 @@ xchat_plugin_deinit ()
 
 	return 1;
 }
-
-
-/*** The End ***/
-/* vim:ts=3:sw=3
- */
