@@ -163,6 +163,7 @@ userlist_button_cb (GtkWidget * button, char *cmd)
 		nicks = malloc (sizeof (char *) * 2);
 		nicks[0] = sess->channel;
 		nicks[1] = NULL;
+		num_sel = 1;
 	} else
 	{
 		/* find number of selected rows */
@@ -255,7 +256,19 @@ menu_quick_item (char *cmd, char *label, GtkWidget * menu, int flags,
 	if (!label)
 		item = gtk_menu_item_new ();
 	else
-		item = gtk_menu_item_new_with_label (label);
+	{
+		if (flags & (1 << 1))
+		{
+			item = gtk_menu_item_new_with_label ("");
+			gtk_label_set_markup (GTK_LABEL (GTK_BIN (item)->child), label);
+		} else
+		{
+			if (flags & (1 << 2))
+				item = gtk_menu_item_new_with_mnemonic (label);
+			else
+				item = gtk_menu_item_new_with_label (label);
+		}
+	}
 	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
 	g_object_set_data (G_OBJECT (item), "u", userdata);
 	if (cmd)
@@ -282,7 +295,7 @@ menu_quick_item_with_callback (void *callback, char *label, GtkWidget * menu,
 }
 
 static GtkWidget *
-menu_quick_sub (char *name, GtkWidget *menu, GtkWidget **sub_item_ret)
+menu_quick_sub (char *name, GtkWidget *menu, GtkWidget **sub_item_ret, int flags, int pos)
 {
 	GtkWidget *sub_menu;
 	GtkWidget *sub_item;
@@ -292,16 +305,20 @@ menu_quick_sub (char *name, GtkWidget *menu, GtkWidget **sub_item_ret)
 
 	/* Code to add a submenu */
 	sub_menu = gtk_menu_new ();
-	sub_item = gtk_menu_item_new_with_label (name);
-	gtk_menu_shell_append (GTK_MENU_SHELL (menu), sub_item);
+	if (flags & 4)
+		sub_item = gtk_menu_item_new_with_mnemonic (name);
+	else
+		sub_item = gtk_menu_item_new_with_label (name);
+	gtk_menu_shell_insert (GTK_MENU_SHELL (menu), sub_item, pos);
 	gtk_widget_show (sub_item);
 	gtk_menu_item_set_submenu (GTK_MENU_ITEM (sub_item), sub_menu);
 
 	if (sub_item_ret)
 		*sub_item_ret = sub_item;
 
-	/* We create a new element in the list */
-	submenu_list = g_slist_prepend (submenu_list, sub_menu);
+	if (flags & 1)
+		/* We create a new element in the list */
+		submenu_list = g_slist_prepend (submenu_list, sub_menu);
 	return sub_menu;
 }
 
@@ -377,7 +394,7 @@ menu_create (GtkWidget *menu, GSList *list, char *target, int check_path)
 		if (!strncasecmp (pop->name, "SUB", 3))
 		{
 			childcount = 0;
-			tempmenu = menu_quick_sub (pop->cmd, tempmenu, &subitem);
+			tempmenu = menu_quick_sub (pop->cmd, tempmenu, &subitem, 1, -1);
 
 		} else if (!strncasecmp (pop->name, "TOGGLE", 6))
 		{
@@ -452,15 +469,16 @@ void
 menu_nickmenu (session *sess, GdkEventButton *event, char *nick, int num_sel)
 {
 	char buf[256];
+	char *fmt;
 	struct User *user;
 	struct away_msg *away;
-	GtkWidget *wid, *submenu, *menu = gtk_menu_new ();
+	GtkWidget *submenu, *menu = gtk_menu_new ();
 
 	if (str_copy)
 		free (str_copy);
 	str_copy = strdup (nick);
 
-	submenu_list = 0;	/* first time though, might not be 0 */
+	submenu_list = 0;	/* first time through, might not be 0 */
 
 	/* more than 1 nick selected? */
 	if (num_sel > 1)
@@ -473,42 +491,44 @@ menu_nickmenu (session *sess, GdkEventButton *event, char *nick, int num_sel)
 		user = find_name_global (current_sess->server, nick);
 		if (user)
 		{
-			submenu = menu_quick_sub (nick, menu, NULL);
+			submenu = menu_quick_sub (nick, menu, NULL, 1, -1);
 
-			snprintf (buf, sizeof (buf), _("User: %s"),
+			/* let the translators tweak this if need be */
+			fmt = _("<tt><b>%-11s</b></tt> %s");
+
+			snprintf (buf, sizeof (buf), fmt, _("User:"),
 						user->hostname ? user->hostname : _("Unknown"));
-			menu_quick_item (0, buf, submenu, 0, 0);
+			menu_quick_item (0, buf, submenu, 2, 0);
 
-			snprintf (buf, sizeof (buf), _("Country: %s"),
+			snprintf (buf, sizeof (buf), fmt, _("Country:"),
 						user->hostname ? country(user->hostname) : _("Unknown"));
-			menu_quick_item (0, buf, submenu, 0, 0);
+			menu_quick_item (0, buf, submenu, 2, 0);
 
-			snprintf (buf, sizeof (buf), _("Realname: %s"),
+			snprintf (buf, sizeof (buf), fmt, _("Real Name:"),
 						user->realname ? user->realname : _("Unknown"));
-			menu_quick_item (0, buf, submenu, 0, 0);
+			menu_quick_item (0, buf, submenu, 2, 0);
 
-			snprintf (buf, sizeof (buf), _("Server: %s"),
+			snprintf (buf, sizeof (buf), fmt, _("Server:"),
 						user->servername ? user->servername : _("Unknown"));
-			menu_quick_item (0, buf, submenu, 0, 0);
+			menu_quick_item (0, buf, submenu, 2, 0);
 
 			if (user->away)
 			{
 				away = find_away_message (current_sess->server, nick);
 				if (away)
 				{
-					snprintf (buf, sizeof (buf), _("Away Msg: %s"),
+					snprintf (buf, sizeof (buf), fmt, _("Away Msg:"),
 								 away->message ? away->message : _("Unknown"));
-					menu_quick_item (0, buf, submenu, 0, 0);
+					menu_quick_item (0, buf, submenu, 2, 0);
 				}
 			}
 
-			snprintf (buf, sizeof (buf), _("Last Msg: %s"),
+			snprintf (buf, sizeof (buf), fmt, _("Last Msg:"),
 						user->lasttalk ? ctime (&(user->lasttalk)) : _("Unknown"));
 			if (user->lasttalk)
 				buf[strlen (buf) - 1] = 0;
-			wid = menu_quick_item (0, buf, submenu, 0, 0);
+			menu_quick_item (0, buf, submenu, 2, 0);
 
-			gtk_label_set_justify (GTK_LABEL (GTK_BIN (wid)->child), GTK_JUSTIFY_LEFT);
 			menu_quick_endsub ();
 			menu_quick_item (0, 0, menu, 1, 0);
 		}
@@ -663,7 +683,7 @@ static void
 menu_chan_cycle (GtkWidget * menu, char *chan)
 {
 	if (current_sess)
-		handle_command (current_sess, "cycle", FALSE);
+		handle_command (current_sess, "CYCLE", FALSE);
 }
 
 static void
@@ -989,7 +1009,7 @@ menu_savedefault (GtkWidget * wid, gpointer none)
 static void
 menu_chanlist (GtkWidget * wid, gpointer none)
 {
-	chanlist_opengui (current_sess->server);
+	chanlist_opengui (current_sess->server, FALSE);
 }
 
 static void
@@ -1341,6 +1361,246 @@ menu_canacaccel (GtkWidget *widget, guint signal_id, gpointer user_data)
 
 #endif
 
+
+/* === STUFF FOR /MENU === */
+
+static GtkMenuItem *
+menu_find_item (GtkWidget *menu, char *name)
+{
+	GList *items = ((GtkMenuShell *) menu)->children;
+	GtkMenuItem *item;
+	GtkWidget *child;
+	const char *labeltext;
+
+	while (items)
+	{
+		item = items->data;
+		child = GTK_BIN (item)->child;
+		if (child)	/* separators arn't labels, skip them */
+		{
+			labeltext = g_object_get_data (G_OBJECT (item), "name");
+			if (!labeltext)
+				labeltext = gtk_label_get_text (GTK_LABEL (child));
+			if (!menu_streq (labeltext, name, 1))
+				return item;
+		} else if (name == NULL)
+		{
+			return item;
+		}
+		items = items->next;
+	}
+
+	return NULL;
+}
+
+static GtkWidget *
+menu_find_path (GtkWidget *menu, char *path)
+{
+	GtkMenuItem *item;
+	char *s;
+	char name[128];
+	int len;
+
+	/* grab the next part of the path */
+	s = strchr (path, '/');
+	len = s - path;
+	if (!s)
+		len = strlen (path);
+	len = MIN (len, sizeof (name) - 1);
+	memcpy (name, path, len);
+	name[len] = 0;
+
+	item = menu_find_item (menu, name);
+	if (!item)
+		return NULL;
+
+	menu = gtk_menu_item_get_submenu (item);
+	if (!menu)
+		return NULL;
+
+	path += len;
+	if (*path == 0)
+		return menu;
+
+	return menu_find_path (menu, path + 1);
+}
+
+static GtkWidget *
+menu_find (GtkWidget *menu, char *path, char *label)
+{
+	GtkWidget *item = NULL;
+
+	if (path[0] != 0)
+		menu = menu_find_path (menu, path);
+	if (menu)
+		item = (GtkWidget *)menu_find_item (menu, label);
+	return item;
+}
+
+static void
+menu_foreach_gui (menu_entry *me, void (*callback) (GtkWidget *, menu_entry *))
+{
+	GSList *list = sess_list;
+	int tabdone = FALSE;
+	session *sess;
+
+	while (list)
+	{
+		sess = list->data;
+		/* do it only once for tab sessions, since they share a GUI */
+		if (!sess->gui->is_tab || !tabdone)
+		{
+			callback (sess->gui->menu, me);
+			if (sess->gui->is_tab)
+				tabdone = TRUE;
+		}
+		list = list->next;
+	}
+}
+
+static void
+menu_update_cb (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item;
+
+	item = menu_find (menu, me->path, me->label);
+	if (item)
+	{
+		gtk_widget_set_sensitive (item, me->enable);
+		/* must do it without triggering the callback */
+		if (GTK_IS_CHECK_MENU_ITEM (item))
+			GTK_CHECK_MENU_ITEM (item)->active = me->state;
+	}
+}
+
+/* toggle state changed via mouse click */
+static void
+menu_toggle_cb (GtkCheckMenuItem *item, menu_entry *me)
+{
+	if (item->active)
+	{
+		me->state = 1;
+		handle_command (current_sess, me->cmd, FALSE);
+	} else
+	{
+		me->state = 0;
+		handle_command (current_sess, me->ucmd, FALSE);
+	}
+
+	/* update the state, incase this was changed via right-click. */
+	/* This will update all other windows and menu bars */
+	menu_foreach_gui (me, menu_update_cb);
+}
+
+static GtkWidget *
+menu_add_toggle (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item = NULL;
+
+	menu = menu_find_path (menu, me->path);
+	if (menu)
+	{
+		item = menu_toggle_item (me->label, menu, menu_toggle_cb, me, me->state);
+		if (me->pos != -1)
+			gtk_menu_reorder_child (GTK_MENU (menu), item, me->pos);
+	}
+	return item;
+}
+
+static GtkWidget *
+menu_add_item (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item = NULL;
+
+	menu = menu_find_path (menu, me->path);
+	if (menu)
+	{
+		item = menu_quick_item (me->cmd, me->label, menu, 4, 0);
+		if (me->pos != -1)
+			gtk_menu_reorder_child (GTK_MENU (menu), item, me->pos);
+	}
+	return item;
+}
+
+static GtkWidget *
+menu_add_sub (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item = NULL;
+
+	if (me->path[0] != 0)
+		menu = menu_find_path (menu, me->path);
+	if (menu)
+		menu_quick_sub (me->label, menu, &item, 4, me->pos);
+	return item;
+}
+
+static void
+menu_del_cb (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item = menu_find (menu, me->path, me->label);
+	if (item)
+		gtk_widget_destroy (item);
+}
+
+static void
+menu_add_cb (GtkWidget *menu, menu_entry *me)
+{
+	GtkWidget *item;
+
+	if (me->ucmd)	/* have unselect-cmd? Must be a toggle item */
+		item = menu_add_toggle (menu, me);
+	else if (me->cmd || !me->label)	/* label=NULL for separators */
+		item = menu_add_item (menu, me);
+	else
+		item = menu_add_sub (menu, me);
+
+	if (item)
+	{
+		gtk_widget_set_sensitive (item, me->enable);
+		if (me->key)
+			gtk_widget_add_accelerator (item, "activate",
+											g_object_get_data (G_OBJECT (menu), "accel"),
+											me->key, me->modifier, GTK_ACCEL_VISIBLE);
+	}
+}
+
+void
+fe_menu_add (menu_entry *me)
+{
+	menu_foreach_gui (me, menu_add_cb);
+}
+
+void
+fe_menu_del (menu_entry *me)
+{
+	menu_foreach_gui (me, menu_del_cb);
+}
+
+void
+fe_menu_update (menu_entry *me)
+{
+	menu_foreach_gui (me, menu_update_cb);
+}
+
+/* used to add custom menus to the right-click menu */
+
+static void
+menu_add_plugin_items (GtkWidget *menu)
+{
+	GSList *list;
+	menu_entry *me;
+
+	list = menu_list;	/* outbound.c */
+	while (list)
+	{
+		me = list->data;
+		menu_add_cb (menu, me);
+		list = list->next;
+	}
+}
+
+/* === END STUFF FOR /MENU === */
+
 GtkWidget *
 menu_create_main (void *accel_group, int bar, int away, int toplevel,
 						GtkWidget **away_item_ret, GtkWidget **user_menu_ret)
@@ -1361,6 +1621,9 @@ menu_create_main (void *accel_group, int bar, int away, int toplevel,
 	else
 		menu_bar = gtk_menu_new ();
 
+	/* /MENU needs to know this later */
+	g_object_set_data (G_OBJECT (menu_bar), "accel", accel_group);
+
 #if GTK_CHECK_VERSION(2,4,0)
 	g_signal_connect (G_OBJECT (menu_bar), "can-activate-accel",
 							G_CALLBACK (menu_canacaccel), 0);
@@ -1407,6 +1670,8 @@ menu_create_main (void *accel_group, int bar, int away, int toplevel,
 			if (mymenu[i].callback == (void *) -1)
 				usermenu = menu;
 			menu_item = gtk_menu_item_new_with_mnemonic (_(mymenu[i].text));
+			/* record the English name for /menu */
+			g_object_set_data (G_OBJECT (menu_item), "name", mymenu[i].text);
 			gtk_menu_shell_append (GTK_MENU_SHELL (menu_bar), menu_item);
 			gtk_widget_show (menu_item);
 			break;
@@ -1467,6 +1732,8 @@ normalitem:
 		case M_MENUSUB:
 			submenu = gtk_menu_new ();
 			item = create_icon_menu (_(mymenu[i].text), mymenu[i].image, TRUE);
+			/* record the English name for /menu */
+			g_object_set_data (G_OBJECT (item), "name", mymenu[i].text);
 			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), submenu);
 			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
 			gtk_widget_show (item);
@@ -1476,7 +1743,10 @@ normalitem:
 			if (!submenu)
 			{
 				if (menu)
+				{
 					gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), menu);
+					menu_add_plugin_items (menu_bar);
+				}
 				if (usermenu)
 					usermenu_create (usermenu);
 				if (bar)
