@@ -95,6 +95,7 @@ static const char *pages[]=
 	"KOI8-R (Cyrillic)",
 	"CP1251 (Cyrillic)",
 	"CP1256 (Arabic)",
+	"CP1257 (Baltic)",
 	"GB18030 (Chinese)",
 	NULL
 };
@@ -609,6 +610,9 @@ servlist_savegui (void)
 static void
 servlist_connectnew_cb (GtkWidget *button, gpointer userdata)
 {
+	if (!selected_net)
+		return;
+
 	if (servlist_savegui () != 0)
 	{
 		gtkutil_simpledialog (_("User name and Real name cannot be left blank."));
@@ -625,6 +629,9 @@ servlist_connectnew_cb (GtkWidget *button, gpointer userdata)
 static void
 servlist_connect_cb (GtkWidget *button, gpointer userdata)
 {
+	if (!selected_net)
+		return;
+
 	if (servlist_savegui () != 0)
 	{
 		gtkutil_simpledialog (_("User name and Real name cannot be left blank."));
@@ -1028,7 +1035,7 @@ servlist_open_edit (GtkWidget *parent, ircnet *net)
 									  &edit_label_user, 0);
 
 	edit_entry_real =
-		servlist_create_entry (table3, _("Real na_me:"), 7, net->real,
+		servlist_create_entry (table3, _("Rea_l name:"), 7, net->real,
 									  &edit_label_real, 0);
 
 	label21 = bold_label (_("Connecting"));
@@ -1421,6 +1428,13 @@ servlist_open_networks (void)
 	gtk_box_pack_start (GTK_BOX (vbox1), hbuttonbox1, FALSE, TRUE, 0);
 	gtk_container_set_border_width (GTK_CONTAINER (hbuttonbox1), 8);
 
+	button_close = gtk_button_new_from_stock ("gtk-close");
+	gtk_widget_show (button_close);
+	g_signal_connect (G_OBJECT (button_close), "clicked",
+							G_CALLBACK (servlist_close_cb), 0);
+	gtk_container_add (GTK_CONTAINER (hbuttonbox1), button_close);
+	GTK_WIDGET_SET_FLAGS (button_close, GTK_CAN_DEFAULT);
+
 	button_connect = gtk_button_new_with_mnemonic (_("C_onnect"));
 	gtk_widget_show (button_connect);
 	g_signal_connect (G_OBJECT (button_connect), "clicked",
@@ -1438,13 +1452,6 @@ servlist_open_networks (void)
 		GTK_WIDGET_SET_FLAGS (button_connectnew, GTK_CAN_DEFAULT);
 	}
 
-	button_close = gtk_button_new_from_stock ("gtk-close");
-	gtk_widget_show (button_close);
-	g_signal_connect (G_OBJECT (button_close), "clicked",
-							G_CALLBACK (servlist_close_cb), 0);
-	gtk_container_add (GTK_CONTAINER (hbuttonbox1), button_close);
-	GTK_WIDGET_SET_FLAGS (button_close, GTK_CAN_DEFAULT);
-
 	gtk_label_set_mnemonic_widget (GTK_LABEL (label3), entry1);
 	gtk_label_set_mnemonic_widget (GTK_LABEL (label6), entry4);
 	gtk_label_set_mnemonic_widget (GTK_LABEL (label7), entry5);
